#include <complex>

#include <BufferedNode.h>
#include <Buffer.h>
#include <Vector.h>
#include <Matrix.h>
#include <Source.h>

#include <Eigen/Dense>
#include <cmath>


using namespace std;
using namespace FD;
using namespace Eigen;

#include <eigen_common.hh>

template <typename Sequence>
struct IndexCompareT {
  IndexCompareT(const Sequence& seq)
    : seq_(seq) { }
  bool operator()(const size_t a, const size_t b) const
  {
    return seq_[a] > seq_[b];
  }
  const Sequence seq_;
};

template <typename Sequence>
IndexCompareT<Sequence>
IndexCompare(const Sequence& seq)
{
  return IndexCompareT<Sequence>(seq);
}

template <typename Sequence>
inline std::vector<size_t> arg_sort(const Sequence& seq)
{
  std::vector<size_t> index(seq.size());
  for (int i = 0; i < index.size(); i++)
    index[i] = i;

  std::sort(index.begin(), index.end(), IndexCompare(seq));

  return index;
}


class SHSFilter;

DECLARE_NODE(SHSFilter);
/*Node
 *
 * @name SHSFilter
 * @category Rescue
 * @description SHSFilter for Robust PCA
 *
 * @input_name INPUT
 * @input_type Matrix<complex<float> >
 * @input_description
 *
 * @input_name INPUT2
 * @input_type Matrix<complex<float> >
 * @input_description
 *
 * @output_name OUTPUT
 * @output_type Vector<complex<float> >
 * @output_description
 *
 * @parameter_name NUM_CHANNEL
 * @parameter_type int
 * @parameter_value 8
 * @parameter_description 
 *
 * @parameter_name SAMPLING_RATE
 * @parameter_type int
 * @parameter_value 16000
 * @parameter_description 
 *
 * @parameter_name LENGTH
 * @parameter_type int
 * @parameter_value 512
 * @parameter_description 
 *
 * @parameter_name NUM_CHANNEL2
 * @parameter_type int
 * @parameter_value 4
 * @parameter_description
 *
END*/

class SHSFilter : public BufferedNode 
{
  int id_input;
  int id_input2;
  int id_output;

  int M, N, F;
  float nyq_freq;

public:
  SHSFilter(string nodeName, ParameterSet params) :
    BufferedNode(nodeName, params)
  {
    id_input  = addInput("INPUT");
    id_input2  = addInput("INPUT2");

    id_output = addOutput("OUTPUT");

    F = dereference_cast<int>(parameters.get("LENGTH")) / 2 + 1;
    nyq_freq = (float)dereference_cast<int>(parameters.get("SAMPLING_RATE")) / 2;

    M = dereference_cast<int>(parameters.get("NUM_CHANNEL"));
    N = dereference_cast<int>(parameters.get("NUM_CHANNEL2"));

    shs_last = VectorXr::Zero(M);

    initParallel();
  }

  VectorXr shs_last;

  static Real non_negative(Real v)
  {
    return v > 0 ? v : 0;
  }

  void calculate(int output_id, int count, Buffer &out)
  {
    // translate RCPtr to Eigen
    RCPtr<FD::Matrix<Complex> > p_input = getInput(id_input, count);
    Eigen::Map<const MatrixXc> input((*p_input)[0], F, M);

    RCPtr<FD::Matrix<Complex> > p_input2 = getInput(id_input2, count);
    Eigen::Map<const MatrixXc> input2((*p_input2)[0], F, M);

    RCPtr<FD::Matrix<Complex> > p_output(new FD::Matrix<Complex>(1, F));
    Eigen::Map<MatrixXc> output((*p_output)[0], F, 1);
    (*(outputs[id_output].buffer))[count] = p_output;

    MatrixXr amp_spec = input.cwiseAbs();
    MatrixXr pwr_spec = input.cwiseAbs();
    
    VectorXr h_list(M);
#pragma omp parallel for
    for (int m=0;m<M;m++) {
      float h_max = 0;
      for (int f0=100; f0<400;f0++) {
	float f_step = f0/nyq_freq*F;
	float h = 0;
	for (float f=f_step;f<F-1;f+=f_step) {
	  h += pwr_spec(round(f), m);
	}
	
	h /= (int)((F - f_step - 2) / f_step + 1);
	if (h > h_max) h_max = h;
      }

      h_list(m) = h_max / pwr_spec.col(m).sum();
    }

    // VectorXr hhh = 0.5 * (h_list + shs_last);
    // shs_last = h_list;
    // vector<size_t> test = arg_sort(hhh);
    vector<size_t> test = arg_sort(h_list);

    // Real hsum = 0;
    // for (int n=0;n<N;n++) hsum += h_list[test[n]];
    // VectorXr outs = VectorXr::Zero(F);
    // for (int n=0;n<N;n++) {
    //   outs += h_list[test[n]] / hsum * amp_spec.col(test[n]);
    // }

    VectorXr outs(F);
#pragma omp parallel for
    for (int f=0;f<F;f++) {
      VectorXr s(N);
      for (int n=0;n<N;n++) {
	s(n) = amp_spec(f, test[n]);
      }

      // outs(f) = s.minCoeff();

      std::sort(s.data(), s.data() + s.size());

      // outs(f) = s(N/2);
      if (N % 2 == 0) {
	outs(f) = (s(N/2-1) + s(N/2)) * 0.5;  // TODO
      	outs(f) = s(N/2-1);
      } else {
      	outs(f) = s(N/2);
      }
    }

    // cout << outs.transpose() << endl;

    output = outs // outs
      .cwiseProduct(input2.col(0))
      .cwiseQuotient(input2.col(0).cwiseAbs());

    // cout << input.col(0).transpose() << endl;
  }
};
